<?php

declare(strict_types=1);

namespace app\models;

use yii;

use app\models\traits\SearchByEdge;

use carono\exchange1c\controllers\ApiController;
use carono\exchange1c\interfaces\DocumentInterface;

use app\models\connection\Dellin;

use Exception;

/**
 * Заказ
 *
 * @see Account Заказчик
 * @see Supply Поставки для заказа
 */
class Order extends Document implements DocumentInterface
{
    use SearchByEdge;

    /**
     * Поставки для записи
     */
    public array|int $supplies;

    /**
     * Имя коллекции
     */
    public static function collectionName(): string
    {
        return 'order';
    }

    /**
     * Свойства
     */
    public function attributes(): array
    {
        return array_merge(
            parent::attributes(),
            [
                'ocid',
                'stts',
                'sync'
            ]
        );
    }

    /**
     * Метки свойств
     */
    public function attributeLabels(): array
    {
        return array_merge(
            parent::attributeLabels(),
            [
                'ocid' => 'Идентификатор 1C',
                'stts' => 'Статус',
                'sync' => 'Статус синхронизации с 1C'
            ]
        );
    }

    /**
     * Правила
     */
    public function rules(): array
    {
        return array_merge(
            parent::rules(),
            [
                [
                    'sync',
                    'boolean',
                    'message' => '{attribute} должен иметь логическое значение'
                ],
                [
                    'sync',
                    'default',
                    'value' => false
                ]
            ]
        );
    }

    /**
     * Подключение к аккаунту
     */
    public function connect(Account $account): ?AccountEdgeOrder
    {
        // Запись ребра: АККАУНТ -> ЗАКАЗ
        return AccountEdgeOrder::write($account->readId(), $this->readId(), 'current') ?? throw new Exception('Не удалось инициализировать ребро: АККАУНТ -> ЗАКАЗ');
    }

    /**
     * Запись товара к заказу
     *
     * $supply = [ Supply $supply, int $amount = 1 ]
     *
     * @param string $supply_id Идентификатор поставки
     * @param string $delivery_type Тип доставки
     * @param int $amount Количество
     * @param Account $trgt Заказчик
     *
     * @return int Количество записанных поставок
     *
     * @todo Создать параметр разделителя для администрации
     */
    public function writeSupply(string $supply_id, string $delivery_type, int $amount = 1, Account $trgt = null): int
    {
        // Инициализация
        $trgt ?? $trgt = yii::$app->user->identity ?? throw new Exception('Не удалось инициализировать заказчика');

        // Проверка корзины
        if (is_null($this->_key)) {
            // Корзина не инициализирована

            // Инициализация корзины
            if (!$this->save()) {
                // Инициализация корзины (активного заказа) не удалась

                throw new Exception('Ошибка при записи заказа в базу данных');
            }

            // Инициализация ребра: АККАУНТ -> ЗАКАЗ
            if (!AccountEdgeOrder::write($trgt->readId(), $this->readId(), 'create')) {
                // Инициализация не удалась

                throw new Exception('Ошибка при записи ребра от аккаунта до заказа в базу данных');
            }
        }

        // Инициализация
        $amount_buffer = 0;

        // Обработка поставок
        for ($i = 0; $i < $amount; $i++) {
            // Создание рёбер соразмерно запросу (добавление нескольких продуктов в корзину)

            // Запись ребра: ЗАКАЗ -> ПОСТАВКА
            if (!$supply_model = Supply::searchById($supply_id) or !$order_edge_supply = OrderEdgeSupply::write($this->readId(), $supply_model->readId(), 'write')) {
                // Поставка не найдена или запись ребра не удалась

                continue;
            } else {
                // Ребро создано (товар подключен к заказу)

                // Обновление счётчика добавленных товаров
                $amount_buffer++;

                // Запись типа доставки
                $order_edge_supply->dlvr = [
                    'type' => $delivery_type
                ];
                $order_edge_supply->update();

                // Запись в журнал
                $this->journal('write', ['target' => $supply_model->readId()]);
            }
        }

        if ($amount_buffer === 0) {
            // Отправка уведомления
            self::notification('Неудачная попытка добавить товар в корзину');
        } else if ($amount_buffer === 1) {
            // Отправка уведомления
            self::notification('Товар ' . $supply_model->catn . ' добавлен в корзину');
        } else {
            // Отправка уведомления
            self::notification('Добавлено ' . $amount_buffer . ' товаров в корзину');
        }

        return $amount_buffer;
    }

    /**
     * Удаление поставки
     *
     * @param Supply|array $supply Товары
     *
     * @return int Количество удалённых рёбер
     */
    public function deleteSupply(Supply|array $supply): int
    {
        // Инициализация
        $amount = 0;

        if ($supply instanceof Supply) {
            // Передана инстанция класса поставки или второй элемент массива не является числом

            // Унификация входных данных
            $supply = [
                $supply->catn => [
                    'auto' => 1
                ]
            ];
        }

        foreach ($supply as $catn => $data) {
            // Перебор целей

            var_dump('ок');

            foreach ($data as $type => $delete_amount) {
                // Перебор данных цели

                var_dump('да');
                if ($supply = Supply::searchByCatn($catn)) {
                    // Поставка найдена

                    $edges = OrderEdgeSupply::searchByVertex($this->readId(), $supply->readId(), limit: $delete_amount, filter: ['order_edge_supply.dlvr.type == \'' . $type . '\'']);

                    for (; count($edges) > $amount; $amount++) {

                        var_dump(count($edges), $amount, $delete_amount);
                        var_dump(PHP_EOL);

                        // Удаление из базы данных
                        $edges[$amount]->delete();

                        // Запись в журнал
                        $this->journal(
                            'delete',
                            [
                                'target' => [
                                    $supply->readId() => $type
                                ]
                            ]
                        );
                    }
                }
            }
        }

        // Генерация вставки текста с типом доставки
        $type = Supply::DeliveryToRussian($type);

        if ($amount === 0) {
            // Отправка уведомления
            self::notification('Не удалось удалить товар из корзины');
        } else if ($amount === 1) {
            // Отправка уведомления
            self::notification("Товар $supply->catn c $type удалён из корзины");
        } else {
            // Отправка уведомления
            self::notification("Удалено $amount товаров из корзины");
        }

        return $amount;
    }

    /**
     * Поиск заказа
     *
     * @todo Привести в порядок
     */
    public static function search(
        Account|string $account = null,
        string $type = 'current',
        string|null $search = null,
        int $limit = 1,
        int $page = 1,
        string|null $select = null,
        bool $supplies = false,
        int|null $from = null,
        int|null $to = null,
        bool $count = false
    ): self|int|array|null {
        // Инициализация аккаунта
        if (empty($account) && isset(yii::$app->user->identity)) {
            // Данные не переданы

            $subquery_where = [
                [
                    'account._id' =>  yii::$app->user->identity->readId()
                ]
            ];
        } else if ($account instanceof Account) {
            // Передан аккаунт

            $subquery_where = [
                [
                    'account._id' =>  $account->readId()
                ]
            ];
        } else if (str_contains($account, '@all')) {
            // Запрос на обработку всех аккаунтов

            $subquery_where = [];
        } else {
            throw new Exception('Не удалось инициализировать пользователя');
        }

        // Инициализация типа заказа
        if (strcasecmp($type, 'all') !== 0) {
            // Если не указан поиск всех заказов

            $subquery_where[] = [
                'account_edge_order.type' => $type
            ];
        }

        // Инициализация сдвига по запрашиваемым данным (пагинация)
        $offset = $limit * ($page - 1);

        // Инициализация фильтрации
        if (isset($from, $to)) {
            // Задан период

            // Инициализация логики
            $foreach = [
                ['edge' => 'account_edge_order'],
                ['jrnl' => 'order.jrnl']
            ];

            // Инициализация фильтра
            $where = "edge._to == order._id && jrnl.action == 'requested' && jrnl.date >= $from && jrnl.date <= $to";
        } else {
            // Ничего не задано

            // Инициализация логики
            $foreach = ['edge' => 'account_edge_order'];

            // Инициализация фильтра
            $where = "edge._to == order._id";
        }

        // Поиск заказов в базе данных
        $return = self::searchByEdge(
            from: 'account',
            to: 'order',
            subquery_where: $subquery_where,
            foreach: $foreach,
            where: $where,
            limit: $limit,
            offset: $offset,
            sort: ['DESC'],
            select: $select,
            direction: 'INBOUND',
            count: !$supplies && $count ? true : false
        );

        if (!$supplies && $count) {
            // Запрошен подсчет заказов

            return $return;
        }

        if ($supplies) {
            // Запрошен поиск поставок

            foreach ($return as &$container) {
                // Перебор заказов

                if ($container instanceof Order) {
                    $buffer = $container;
                } else {
                    // Инициализация настроек
                    $config = $container['order'];
                    unset($config['_id'], $config['_rev'], $config['_id']);

                    // Инициализация буфера
                    $buffer = new Order($config);
                }

                // Чтение полного содержания
                $container['supplies'] = $buffer->content($limit, $page, $search, count: $count);

                if ($count) {
                    // Запрошен подсчет поставок

                    return $container['supplies'];
                }
            }
        }

        return $limit === 1 ? $return[0] ?? null : $return;
    }

    // /**
    //  * Объеденить дубликаты поставок
    //  *
    //  * @return array Данные заказа
    //  */
    // public function mergeDuplicateSupplies(array $order): array {
    //     foreach ($orders as $order) {

    //     }
    // }

    /**
     * Поиск содержимого заказа
     *
     * @todo В будущем возможно заказ не только поставок реализовать
     * Переписать реестр и проверку на дубликаты, не понимаю как они работают
     */
    public function content(int $limit = 1, int $page = 1, string|null $search = null, bool $count = false): Supply|int|array|null
    {
        // Генерация сдвига по запрашиваемым данным (пагинация)
        $offset = $limit * ($page - 1);

        if (!empty($search)) {
            // Передан поиск по продуктам в заказах

            // Добавить ограничитель

            $limit = 9999;
        }

        // Поиск по рёбрам: ЗАКАЗ -> ПОСТАВКА
        $connections = Supply::searchByEdge(
            from: 'order',
            to: 'supply',
            edge: 'order_edge_supply',
            subquery_where: [
                [
                    'order._id' => $this->readId()
                ]
            ],
            foreach: ['edge' => 'order_edge_supply'],
            where: 'edge._to == supply._id',
            limit: $limit,
            offset: $offset,
            filterStart: ['catn' => $search],
            direction: 'INBOUND',
            select: '{supply, order_edge_supply}',
            count: $count
        );

        if ($count) {
            // Запрошен подсчет

            return $connections;
        }

        // Рекурсивная фильтрации соединений
        recursive_connections_filtration:

        // Инициализация статуса об изменении списка поставок
        $changed = false;

        // Инициализация глобального буфера поставок
        $buffer_global = [];

        // Разделение поставок по типам доставки
        // foreach ($connections as &$connection) {
        //     // Перебор поставок

        //     if (count($connection['order_edge_supply'] ?? []) > 1) {
        //         // Поставок более чем одна

        //         // Инициализация (выбор значения для фильтрации)
        //         $target_dlvr_type = reset($connection['order_edge_supply'])['dlvr']['type'] ?? 'auto';

        //         // Инициализация буфера поставок
        //         $buffer = $connection;
        //         unset($buffer['order_edge_supply']);

        //         foreach ($connection['order_edge_supply'] as $key => &$order_edge_supply) {
        //             // Перебор рёбер: ЗАКАЗ -> ПОСТАВКА

        //             // if ((empty($order_edge_supply['dlvr']['type']) && $target_dlvr_type === 'auto')
        //             //     || (isset($order_edge_supply['dlvr']['type']) && $order_edge_supply['dlvr']['type'] !== $target_dlvr_type)
        //             // ) {
        //             if ((isset($order_edge_supply['dlvr']['type']) && $order_edge_supply['dlvr']['type'] === $target_dlvr_type)
        //                 || (empty($order_edge_supply['dlvr']['type']) && $target_dlvr_type === 'auto')
        //             ) {
        //                 // Тип доставки для одной поставки отличается

        //                 // Запись в буфер
        //                 $buffer['order_edge_supply'][$key] = $order_edge_supply;

        //                 // Удаление из списка
        //                 unset($connection['order_edge_supply'][$key]);

        //                 // Перезапись статуса
        //                 $changed = true;
        //             }
        //         }

        //         // Запись из буфера в глобальный буфер
        //         $buffer_global[] = $buffer;
        //     }
        // }

        // $connections = array_merge($connections, $buffer_global);

        // if ($changed) goto recursive_connections_filtration;

        // Инициализация реестра дубликатов
        $registry = [];

        // Подсчёт и перестройка массива для очистки от дубликатов
        foreach ($connections as $key => &$connection) {
            // Перебор поставок

            if (in_array([
                'catn' => $connection['supply']['catn'],
                'type' => $type = reset($connection['order_edge_supply'])['dlvr']['type'] ?? 'auto'
            ], $registry)) {
                // Если данная поставка найдена в реестре

                // Удаление
                unset($connections[$key]);

                // Пропуск итерации
                continue;
            }

            // Повторный перебор для поиска дубликатов
            foreach ($connections as &$connection_for_check) {
                if ($connection == $connection_for_check) {
                    // Найден дубликат

                    // Запись в реестр
                    $registry[$key] = [
                        'catn' => $connection_for_check['supply']['catn'],
                        'type' => $type
                    ];
                }
            }

            // Инициализация счетчиков
            $connection['amount'] = [
                'auto' => 0,
                'avia' => 0
            ];

            // Подсчет количества поставок
            foreach ($connection['order_edge_supply'] as $edge) {
                // Перебор связанных поставок

                if ($edge['dlvr']['type'] === 'auto') ++$connection['amount']['auto'];
                if ($edge['dlvr']['type'] === 'avia') ++$connection['amount']['avia'];
            }
        }

        // Инициализация дополнительных данных
        foreach ($connections as $key => &$connection) {
            // Перебор поставок

            // Чтение стоимости
            $cost = Supply::readCostById($connection['supply']['_id']);

            if (empty($cost) || $cost['ЦенаЗаЕдиницу'] < 1) {
                // Если стоимость не найдена или равна нулю (явная ошибка)

                // Удаление из базы данных
                $this->deleteSupply($connection['supply']['_id']);

                ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Добавить уведомление об ошибочном товаре

                // Удаление из списка
                unset($connections[$key]);

                // Пропуск итерации
                continue;
            }

            // Поиск ребра до аккаунта
            $connection['account'] = Account::searchBySupplyId($connection['supply']['_id']);

            // Поиск привязанного товара
            $connection['product'] = Product::searchBySupplyId($connection['supply']['_id']);

            // if (empty(reset($connection['order_edge_supply'])['dlvr']['type']) || reset($connection['order_edge_supply'])['dlvr']['type'] === 'auto') {
            //     // Доставка автоматическая

            try {
                // Инициализация данных геолокации
                try {
                    $from = (int) $connection['account']['opts']['delivery_from_terminal'] ?? empty(Settings::search()->delivery_from_default) ? 36 : (int) Settings::search()->delivery_from_default;
                } catch (Exception $e) {
                    $from = empty(Settings::search()->delivery_from_default) ? 36 : (int) Settings::search()->delivery_from_default;
                }

                try {
                    $to = (int) yii::$app->user->identity->opts['delivery_to_terminal'] ?? 36;
                } catch (Exception $e) {
                    $to = 36;
                }

                if ($buffer_connection = $connection['product']['bffr']["$from-$to"] ?? false) {
                    // Найдены данные доставки в буфере

                    if (time() < $buffer_connection['expires']) {
                        // Срок хранения не превышен, информация актуальна

                        // Запись в буфер вывода
                        $connection['delivery']['auto'] = $buffer_connection['data'];
                    }
                } else {
                    // Инициализация инстанции продукта в базе данных
                    $product = Product::searchByCatn($connection['product']['catn']);

                    // Инициализация доставки Dellin (автоматическая)
                    $product->bffr = ($product->bffr ?? []) + [
                        "$from-$to" => [
                            'data' => $connection['delivery']['auto'] = Dellin::calcDeliveryAdvanced(
                                $from,
                                $to,
                                (int) ($connection['product']['wght'] ?? 0),
                                (int) ($connection['product']['dmns']['x'] ?? 0),
                                (int) ($connection['product']['dmns']['y'] ?? 0),
                                (int) ($connection['product']['dmns']['z'] ?? 0),
                                count($connection['order_edge_supply'])
                            ),
                            'expires' => time() + 86400
                        ]
                    ];

                    // Отправка в базу данных
                    $product->update();
                }
            } catch (Exception $e) {
                $connection['delivery']['auto']['error'] = true;

                // echo '<pre>';
                // var_dump($e->getMessage());
                // var_dump($e->getTrace());
                // var_dump($e->getFile());
                // die;

                // var_dump(json_decode($e->getMessage(), true)['errors']); die;
            }

            // Запись цены (цена поставки + цена доставки + наша наценка)
            $connection['cost']['auto'] = ($cost['ЦенаЗаЕдиницу'] ?? $connection['supply']->onec['Цены']['Цена']['ЦенаЗаЕдиницу']) + ($connection['delivery']['auto']['price']['all'] ?? $connection['delivery']['auto']['price']['one'] ?? 0) + ($settings['increase'] ?? 0) ?? 0;
            // } else {
            // Доставка самолётом

            try {
                // Инициализация данных геолокации
                try {
                    $from = (int) $connection['account']['opts']['delivery_from_terminal'] ?? empty(Settings::search()->delivery_from_default) ? 36 : (int) Settings::search()->delivery_from_default;
                } catch (Exception $e) {
                    $from = empty(Settings::search()->delivery_from_default) ? 36 : (int) Settings::search()->delivery_from_default;
                }

                try {
                    $to = (int) yii::$app->user->identity->opts['delivery_to_terminal'] ?? 36;
                } catch (Exception $e) {
                    $to = 36;
                }

                if ($buffer_connection = $connection['product']['bffr']["$from-$to-avia"] ?? false) {
                    // Найдены данные доставки в буфере

                    if (time() < $buffer_connection['expires']) {
                        // Срок хранения не превышен, информация актуальна

                        // Запись в буфер вывода
                        $connection['delivery']['avia'] = $buffer_connection['data'];
                    }
                } else {
                    // Инициализация инстанции продукта в базе данных
                    $product = Product::searchByCatn($connection['product']['catn']);

                    // Инициализация доставки Dellin (автоматическая)
                    $product->bffr = ($product->bffr ?? []) + [
                        "$from-$to-avia" => [
                            'data' => $connection['delivery']['avia'] = Dellin::calcDeliveryAdvanced(
                                $from,
                                $to,
                                (int) ($connection['product']['wght'] ?? 0),
                                (int) ($connection['product']['dmns']['x'] ?? 0),
                                (int) ($connection['product']['dmns']['y'] ?? 0),
                                (int) ($connection['product']['dmns']['z'] ?? 0),
                                count($connection['order_edge_supply']),
                                avia: true
                            ),
                            'expires' => time() + 86400
                        ]
                    ];

                    // Отправка в базу данных
                    $product->update();
                }
            } catch (Exception $e) {
                $connection['delivery']['avia']['error'] = true;

                // var_dump($e->getMessage());
                // var_dump($e->getTrace());
                // var_dump($e->getFile());
                // die;

                // var_dump(json_decode($e->getMessage(), true)['errors']); die;
            }

            // Запись цены (цена поставки + цена доставки + наша наценка)
            $connection['cost']['avia'] = ($cost['ЦенаЗаЕдиницу'] ?? $connection['supply']->onec['Цены']['Цена']['ЦенаЗаЕдиницу']) + ($connection['delivery']['avia']['price']['all'] ?? $connection['delivery']['avia']['price']['one'] ?? 0) + ($settings['increase'] ?? 0) ?? 0;
            // }

            // Запись валюты
            $connection['currency'] = $cost['Валюта'];
        }

        return $connections;
    }

    /**
     * Проверка на то, что все поставки подтверждены
     *
     * @param array $order_edge_supply Поставки
     *
     * @return bool Статус подтверждения всех поставок (true если все и false если хотя бы одна из них не подтверждена)
     */
    public static function checkSuppliesStts(array $order_edge_supply): bool
    {
        foreach ($order_edge_supply as $edge) {
            // Перебор поставок

            if (empty($edge['stts']) || $edge['stts'] !== 'accepted') {
                // Найдена неподтверждённая поставка

                return false;
            }
        }

        return true;
    }

    /**
     * @return DocumentInterface[]
     */
    public static function findDocuments1c(): ?array
    {
        // yii::$app->on(ApiController::EVENT_AFTER_EXPORT_ORDERS, self::afterExport1c());

        $orders = self::searchByEdge(
            from: 'account',
            to: 'order',
            edge: 'account_edge_order',
            direction: 'INBOUND',
            subquery_where: 'account_edge_order.type == "processed"',
            where: ['sync' => false]
        );

        foreach ($orders as &$order) {
            // Перебор заказов

            // Запись о том, что синхронизация проведена
            $order->sync = true;

            $order->update();
        }

        return $orders;
    }

    /**
     * @return OfferInterface[]
     */
    public function getOffers1c(): mixed
    {
        // Инициализация
        $supplies = [];

        foreach ($this->jrnl as $key => $jrnl) {
            // Перебор журнала

            // if (isset($supplies[$key]) && $supplies[$key]['id'] !== $jrnl['target']) {
            //     // Запись уже существует и идентификаторы не совпадают

            //     $key .= '_du'

            //     // Реинициализация
            //     $supplies[$key]['id'] = $jrnl['target'];
            // } else {
            //     // Инициализация
            // }


            if (($jrnl['action'] ?? null) === 'write') {
                // Найдено событие записи товара к заказу

                $supplies[$key]['id'] = $jrnl['target'];

                $supplies[$key]['amount'] ?? $supplies[$key]['amount'] = 0;

                ++$supplies[$key]['amount'];
            } else if (($jrnl['action'] ?? null) === 'delete') {
                // Найдено событие удаления товара из заказа

                $supplies[$key]['id'] = $jrnl['target'];

                $supplies[$key]['amount'] ?? $supplies[$key]['amount'] = 0;

                --$supplies[$key]['amount'];
            }
        }

        // file_put_contents('supplies.txt', print_r($supplies, true));

        if (count($supplies) > 0) {
            // Поставки были записаны

            // Инициализация
            $supplies_buffer = [];

            foreach ($supplies as $id => $supply) {
                // Перебор поставок

                if ($supply['amount'] < 1) {
                    continue;
                }

                if ($response = Supply::searchById($supply['id'])) {
                    // Поставка найдена в базе данных

                    $supplies_buffer[] = $response;
                }
            }

            return $supplies_buffer;
        }
        // file_put_contents('AAAAAAAAAAAAAAAAAAA.txt', print_r(1, true));

        return [];
    }

    /**
     * Неизвестно для чего
     */
    public function getRequisites1c(): void
    {
        // return true;
    }

    /**
     * Получить контрагента (пользователя от лица которого происходит операция)
     *
     * @return PartnerInterface
     */
    public function getPartner1c(): Account
    {
        return yii::$app->user->identity ?? throw new Exception('Не удалось идентифицировать пользователя');
    }

    public function getExportFields1c($context = null)
    {
        return [];
    }

    /**
     * Возвращаем имя поля в базе данных, в котором хранится ID из 1с
     *
     * @return string
     */
    public static function getIdFieldName1c()
    {
        return 'ocid';
    }

    public function setRaw1cData($cml, $object): void
    {
    }

    protected static function afterExport1c(): void
    {
        file_put_contents('afterExport1c.txt', print_r(1, true));
    }

    /**
     * Отправка уведомления
     */
    public static function notification(string $text, string|null $type = Notification::TYPE_NOTICE): Notification|array|null
    {
        // Отправка
        return Notification::_write($text, type: $type);
    }

    /**
     * Посчитать количство заказов или количество поставок в заказах
     *
     * @param int $limit Ограничение
     * @param bool $supplies Считать поставки в активном заказе (иначе - заказы)
     *
     * @return int Количество
     */
    public static function count(int $limit = 500, bool $supplies = false): int
    {
        return self::search(supplies: $supplies, type: $supplies ? 'current' : 'all', limit: $limit, count: true);
    }
}
